(*
This file is part of the Kind verifier

* Copyright (c) 2007-2009 by the Board of Trustees of the University of Iowa, 
* here after designated as the Copyright Holder.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Iowa, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)

open Types
open Exceptions

(* a "safe" version of Hashtbl.find that raises a more informative exception *)
let safe_find ht x s = 
  try
    Hashtbl.find ht x
  with  Not_found -> 
    raise (IdNotFound s)

let reverse_varinfo = (Hashtbl.create 100:(string,idtype)Hashtbl.t)
let varid_name_add name id = Hashtbl.replace reverse_varinfo name id
let varid_lookup name = safe_find reverse_varinfo name "varid_lookup"


(* this renames long varnames internally so we don't get odd hashtable clashes *)
let renaming_counter = new Counter.counter

let sym_renaming_table = (Hashtbl.create 100:(string,string)Hashtbl.t)
(* the normal hashtbl does not behave as expected with long strings *)
  module LongStringHash = 
    Hashtbl.Make (
      struct
        type t = string
        let equal x y = ((compare (x:string) (y:string)) = 0)
        let hash x = Hashtbl.hash_param 200 200 (x:string)
      end
    )

let sym_truenaming_table = LongStringHash.create 1

let rename_sym s =
  if (!Flags.rename_long_vars) && ((String.length s) > Flags.long_vars_length) then 
    begin
      try
        LongStringHash.find sym_truenaming_table s
      with Not_found ->
        let s1 = "___z"^(string_of_int renaming_counter#next)^"z___" in
        Hashtbl.replace sym_renaming_table s1 s;
        LongStringHash.replace sym_truenaming_table s s1;
        s1
    end
  else
    s

let resolve_var_name n =
  try
    Hashtbl.find sym_renaming_table n
  with Not_found -> n

(* shared bindings *)
(* variable information, referenced by id number *)
(* node, name, type, variable class *)
let varinfo = 
  (Hashtbl.create 100:(int,node_id_t*string*lustre_type*varclass)Hashtbl.t)
let update_varinfo id (node,var,t,c) =
  let var' = rename_sym (resolve_var_name var) in
  Hashtbl.replace varinfo id (node,var',t,c)
let get_varinfo () = varinfo
let is_input_var id =
  let (_,_,_,c) = safe_find varinfo id ("is_input_var "^(string_of_int id)) in
  c = INPUT
let is_local_var id =
  let (_,_,_,c) = safe_find varinfo id ("is_local_var "^(string_of_int id)) in
  c = LOCAL || c = ABSTRACT
let is_output_var id =
  let (_,_,_,c) = safe_find varinfo id ("is_output_var "^(string_of_int id)) in
  c = OUTPUT
let string_of_varinfo id =
   let (n,v,t,c) = 
     safe_find varinfo id ("string_of_varinfo "^(string_of_int id)) 
   in
   (string_of_int id)^":("^v^")"
let safe_find_varinfo x label = safe_find varinfo x label
let get_varinfo_name x = 
  let (_,v,_,_) = safe_find varinfo x "get_varinfo_name" in
  v


(* keeps track of which variables are actually used and at what position(s) *)
(* also keeps a il_expression reference *)
let used_vars = (Hashtbl.create 100:(idtype,used_vars_type option)Hashtbl.t)
let get_used_vars () = used_vars
let find_used_var = Hashtbl.find used_vars
(* when we encounter a "used" var, update at what position it is referenced *)
let update_used_vars id depth e = 
assert (depth >= 0);
  try
    let entry = Hashtbl.find used_vars id in
    if depth > 0 then
      match entry with
         None -> Hashtbl.replace used_vars id (Some {ex=e;depths=[depth]})
       | Some r -> if not (List.mem depth r.depths) then
                      r.depths <- depth::r.depths
  with Not_found ->
    if depth = 0 then
      Hashtbl.replace used_vars id None
    else
      Hashtbl.replace used_vars id (Some {ex=e;depths=[depth]})

let var_is_stateful id =
  try
    match find_used_var id with
        Some _ -> true
      | None -> false
  with Not_found -> false


(* keep track of the shortcut functions actually used *)
(* used to determine which cvc3 functions to define *)
let shortcuts_used = (Hashtbl.create 10: (string,bool)Hashtbl.t)
let set_shortcut_used x = Hashtbl.replace shortcuts_used x true
let get_shortcut_used x = Hashtbl.mem shortcuts_used x


(* substitutions *)
(* keeps track of number of rhs occurrances of vars, for inlining complexity *)
let var_count_table = (Hashtbl.create 10:(idtype,int)Hashtbl.t)
let add_var_count id num =
  try
    let n = Hashtbl.find var_count_table id in
    Hashtbl.replace var_count_table id (n+num)
  with Not_found ->
    Hashtbl.add var_count_table id num
let get_var_count id =
  try
    (Hashtbl.find var_count_table id)-1
  with Not_found -> 0
(* this is in a table because properties are not tied to other definitions,
even though they may be in non-main nodes.  If this is the case, we may
have a property variable that does not actually show up anywhere else in the
formula. *)
(* if a variable is not in the table, it does not need to be substituted *)
(* inputs replace parameters, and results replace outputs *)
let number_variables_inlined = ref 0
let get_number_variables_inlined () = !number_variables_inlined
let substitution_table = (Hashtbl.create 10:(idtype,idtype)Hashtbl.t)
let rec resolve_substitution x =
  try
    resolve_substitution (Hashtbl.find substitution_table x)
  with Not_found -> x
let add_substitution x y =
  incr number_variables_inlined;
  add_var_count y ((get_var_count x) - 1);
  Hashtbl.replace substitution_table (resolve_substitution x)
    (resolve_substitution y)

(* node definitions, referenced by node name *)
let node_inputs = (Hashtbl.create 10:(node_id_t,(idtype*lustre_type) list)Hashtbl.t)

let get_node_inputs nid = safe_find node_inputs nid "get_node_inputs"
let add_node_inputs = Hashtbl.replace node_inputs

let node_outputs = (Hashtbl.create 10:(node_id_t,(idtype*lustre_type) list)Hashtbl.t)

let get_node_outputs nid = safe_find node_outputs nid "get_node_outputs"
let add_node_outputs = Hashtbl.replace node_outputs

let node_locals = (Hashtbl.create 10:(node_id_t,(idtype*lustre_type) list)Hashtbl.t)

let get_node_locals = Hashtbl.find node_locals
let add_node_locals = Hashtbl.replace node_locals

(* list of other nodes called: (node name, line, col) *)
let node_defs = (Hashtbl.create 10:(node_id_t,typed_stream list)Hashtbl.t)

let get_node_defs nid = safe_find node_defs nid "get_node_defs"
let add_node_defs = Hashtbl.replace node_defs

(* from node id to node name *)
let nodenamehash = (Hashtbl.create 100: (int,string)Hashtbl.t)

let get_nodename_table () = nodenamehash
let get_nodename nid = safe_find nodenamehash nid "get_nodename"
let add_nodename = Hashtbl.replace nodenamehash



