(*
This file is part of the Kind verifier

* Copyright (c) 2007-2009 by the Board of Trustees of the University of Iowa, 
* here after designated as the Copyright Holder.
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the University of Iowa, nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER ''AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)

(** Functions to reference a number of global databases. *)

(** A "safe" version of Hashtbl.find that raises a more informative exception {!Exceptions.IdNotFound} *)
val safe_find : ('a,'b)Hashtbl.t -> 'a -> string -> 'b 

(** Associate a string with a variable id *)
val varid_name_add : string -> Types.idtype -> unit 

(** Given a variable id, return the associated string *)
val varid_lookup : string -> Types.idtype


(** Ocaml hashtable have a limit on how they compare keys, so in order to
support longer identifier names in the lustre source and also to avoid
any namespace clashes, we rename identifiers longer than {!Flags.long_vars_length} as soon as we encounter them. This will be consistent up to about 100 characters. *)
val rename_sym : string -> string

(** Given an internal string name, return the original version. *)
val resolve_var_name : string -> string 


(* shared bindings *)
(* variable information, referenced by id number *)
(* node, name, type, variable class *)

(** The [varinfo] table is the main symbol table, containing a variable's node of origin, string representation, type, and classification (input/output/etc). *)

(** Update the [varinfo] table with a new entry.  *)
val update_varinfo : Types.idtype -> 
  Types.node_id_t * string * Types.lustre_type * Types.varclass -> unit

(** Return the [varinfo] table *)
val get_varinfo : unit -> Types.varinfo_table

(** True if input.  Raises {!Exceptions.IdNotFound} on invalid query. *)
val is_input_var : Types.idtype -> bool

(** True if local.  Raises {!Exceptions.IdNotFound} on invalid query. *)
val is_local_var : Types.idtype -> bool

(** True if output.  Raises {!Exceptions.IdNotFound} on invalid query. *)
val is_output_var : Types.idtype -> bool

(** [safe_find_varinfo x label] Returns the full varinfo entry of a variable [x]; [label] is a descriptive string (for debugging). Raises {!Exceptions.IdNotFound} on invalid query. *)
val safe_find_varinfo : Types.idtype -> string ->
  Types.node_id_t * string * Types.lustre_type * Types.varclass

(** Returns the string name of a variable. Raises {!Exceptions.IdNotFound} on invalid query. *)
val get_varinfo_name : Types.idtype -> string


(** The [used_vars] table keeps track of which variables are actually used and at what position(s), along with a {!Types.il_expression} reference used when generating path compression constraints (see {!Types.used_vars_type}).  Each entry is an option, with [None] indicating the variable is not stateful. *)

(** Returns the [used_vars] table. Generally then iterated over. *)
val get_used_vars : unit -> (Types.idtype,Types.used_vars_type option)Hashtbl.t

(** Return just one entry.  Raises [Not_found] on bad query. *)
val find_used_var : Types.idtype -> Types.used_vars_type option

(** When we encounter a "used" var, update at what position it is referenced *)
val update_used_vars : Types.idtype -> int -> Types.il_expression -> unit

(** Returns true if a variable has been determined to be stateful.  Otherwise returns false. *)
val var_is_stateful : Types.idtype -> bool


(** The solver may identify certain functions with macro equivalents
(if, for example, a given function is not built in),
or "shortcuts".  The table [shortcuts_used] keeps track of these, so
Kind only needs to define those shortcuts that are actually relevant to 
the lustre program. *)

(** Register a shortcut. *)
val set_shortcut_used : string -> unit

(** Return true if a shortcut has been registered. *)
val get_shortcut_used : string -> bool


(** Kind keeps track of the number of variable occurrences for inlining heuristics. *)

(** Variable x has occurred y (more) times on the RHS of an equation.  This is cumulative. *)
val add_var_count : Types.idtype -> int -> unit

(** Return total number of RHS occurrences of a given variable.  *)
val get_var_count : Types.idtype -> int

(** Stats on total number inlined so far *)
val get_number_variables_inlined : unit -> int

(** Kind handles variable unification via a [substitution_table]. *)

(** Given a variable id, return its unification class representative id *)
val resolve_substitution : Types.idtype -> Types.idtype

(** [add_substitution x y] marks [y] as [x]'s representative ([x] will be replaces throughout the program).  *)
val add_substitution : Types.idtype -> Types.idtype -> unit

(** Given a node id, return a list of its input parameters *)
val get_node_inputs : Types.node_id_t -> 
  (Types.idtype * Types.lustre_type) list

(** Add an entry to the [node_inputs] table *)
val add_node_inputs : Types.node_id_t -> 
  (Types.idtype * Types.lustre_type) list -> unit


(** Given a node id, return a list of its output variables *)
val get_node_outputs : Types.node_id_t -> 
  (Types.idtype * Types.lustre_type) list


(** Replace an entry to the [node_outputs] table *)
val add_node_outputs : Types.node_id_t -> 
  (Types.idtype * Types.lustre_type) list -> unit


(** Given a node id, return a list of its input parameters *)
val get_node_locals  : Types.node_id_t ->
  (Types.idtype * Types.lustre_type) list

(** Replace an entry to the [node_locals] table *)
val add_node_locals : Types.node_id_t ->
  (Types.idtype * Types.lustre_type) list -> unit


(** Given a node id, return a list of all its [S_DEFS] *)
val get_node_defs : Types.node_id_t -> Types.typed_stream list 

(** Replace an entry to the [node_defs] table *)
val add_node_defs : Types.node_id_t -> Types.typed_stream list -> unit

(** Returns the [nodename] table *)
val get_nodename_table : unit -> (Types.node_id_t,string)Hashtbl.t

(** Given a node it, return its name *)
val get_nodename : Types.node_id_t -> string

(** Replace an entry to the [nodename] table *)
val add_nodename : Types.node_id_t -> string -> unit



